//   Copyright 2012 Georg-August-Universität Göttingen, Germany
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.

package de.ugoe.cs.autoquest.tasktrees.treeimpl;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

import de.ugoe.cs.autoquest.tasktrees.treeifc.ITask;
import de.ugoe.cs.autoquest.tasktrees.treeifc.ITaskInstance;
import de.ugoe.cs.autoquest.tasktrees.treeifc.ITaskVisitor;

/**
 * <p>
 * this is the default implementation of the interface {@link ITask}. It
 * does not do anything fancy except implementing the interface.
 * </p>
 * 
 * @author Patrick Harms
 */
class Task implements ITask {

    /**
     * <p>
     * default serial version UID
     * </p>
     */
    private static final long serialVersionUID = 1L;

    /**
     * <p>
     * used as a counter to generate new ids for each newly created task. May overflow.
     * </p>
     */
    private static int temporalId = 0;

    /**
     * <p>
     * the id of the task (unique throughout the system as long as {@link #temporalId} does not
     * overflow.
     * </p>
     */
    private int id;
    
    /**
     * <p>
     * a human readable type of the task (used for visualization purposes)
     * </p>
     */
    private String type;

    /**
     * <p>
     * a human readable description of the task
     * </p>
     */
    private String description;
    
    /**
     * <p>
     * the instances of this task
     * </p>
     */
    private Collection<ITaskInstance> instances = new HashSet<ITaskInstance>();

    /**
     * <p>
     * constructs a new task with a new id. The id is generated using the {@link #getNewId()}
     * method
     * </p>
     * 
     * @param type the human readable type of the task
     * 
     * @throws IllegalArgumentException in the case the provided type is null
     */
    Task(String type) {
        this.id = getNewId();
        this.type = type;
        
        if (type == null) {
            throw new IllegalArgumentException("type must not be null");
        }
    }

    /**
     * <p>
     * creates a new id for a task using {@link #temporalId} by incrementing it an returning its
     * current value. Resets the counter if {@link Integer.MAX_VALUE} is reached.
     * </p>
     * 
     * @return a new unique id for a task as long as {@link #temporalId} does not overflow
     */
    private static synchronized int getNewId() {
        if (temporalId == Integer.MAX_VALUE) {
            temporalId = 0;
        }

        return temporalId++;
    }

    /* (non-Javadoc)
     * @see de.ugoe.cs.autoquest.tasktrees.treeifc.ITask#geId()
     */
    @Override
    public int getId() {
        return id;
    }

    /* (non-Javadoc)
     * @see de.ugoe.cs.autoquest.tasktrees.treeifc.ITask#getType()
     */
    @Override
    public String getType() {
        return type;
    }

    /* (non-Javadoc)
     * @see de.ugoe.cs.autoquest.tasktrees.treeifc.ITask#getDescription()
     */
    @Override
    public String getDescription() {
        return description;
    }

    /* (non-Javadoc)
     * @see de.ugoe.cs.autoquest.tasktrees.treeifc.ITask#getInstances()
     */
    @Override
    public Collection<ITaskInstance> getInstances() {
        return Collections.unmodifiableCollection(instances);
    }

    /* (non-Javadoc)
     * @see de.ugoe.cs.autoquest.tasktrees.treeifc.ITask#equals(ITask)
     */
    @Override
    public final boolean equals(ITask task) {
        // tasks are only equal if they are identical or if they have the same id
        // (may happen, if they are cloned)
        return (this == task) || (this.hashCode() == task.hashCode());
    }

    /* (non-Javadoc)
     * @see de.ugoe.cs.autoquest.tasktrees.treeifc.ITask#hashCode()
     */
    @Override
    public synchronized int hashCode() {
        return id;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public synchronized String toString() {
        StringBuffer result = new StringBuffer();
        result.append(type);
        result.append(" #");
        result.append(id);
        
        if (description != null) {
            result.append(" (");
            result.append(description);
            result.append(')');
        }
        
        return result.toString();
    }

    /* (non-Javadoc)
     * @see de.ugoe.cs.autoquest.tasktrees.treeifc.ITask#clone()
     */
    @Override
    public synchronized ITask clone() {
        Task clone = null;
        try {
            clone = (Task) super.clone();
        }
        catch (CloneNotSupportedException e) {
            // this should never happen. Therefore simply dump the exception
            e.printStackTrace();
        }

        return clone;
    }

    /**
     * <p>
     * internally used to set the human readable description of the task
     * </p>
     * 
     * @param description the new human readable description of the task
     */
    void setDescription(String description) {
        this.description = description;
    }

    /**
     * <p>
     * internally used to remove an instance from this task
     * </p>
     * 
     * @param instance the instance to be removed from this task
     */
    void removeInstance(ITaskInstance instance) {
        this.instances.remove(instance);
    }

    /**
     * <p>
     * internally used to add an instance to this task
     * </p>
     * 
     * @param instance the instance belonging to this task
     */
    void addInstance(ITaskInstance instance) {
        this.instances.add(instance);
    }
    
    /* (non-Javadoc)
     * @see de.ugoe.cs.autoquest.tasktrees.treeifc.ITask#accept(ITaskVisitor)
     */
    @Override
    public void accept(ITaskVisitor visitor) {
        visitor.visit(this);
    }

}
